/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* Overveiw of the network TODO update
 *         nT
 *          |
 *nR1-nR1d-nTd-nR2d-nR2
 * 
 */

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/csma-module.h"
#include "ns3/arp-cache.h"
#include "ns3/ipv4-static-routing-helper.h"
#include "ns3/netanim-module.h"
#include "ns3/flow-monitor-helper.h"

// Normal C++ stuff
#include <iostream>
#include <fstream>
#include <string>

using namespace ns3;
using std::cout;

NS_LOG_COMPONENT_DEFINE ("NetworkDiodeTest");

int
main (int argc, char *argv[])
{
  uint32_t nPackets = 100;
  uint32_t nClients = 10;

  CommandLine cmd;
  cmd.AddValue("nPackets", "Number of packets to echo", nPackets);
  cmd.AddValue("nClients", "Number of extra clients to run", nClients);
  cmd.Parse (argc, argv);
  
  Time::SetResolution (Time::NS);
  LogComponentEnable ("UdpEchoClientApplication", LOG_LEVEL_INFO);
  LogComponentEnable ("UdpEchoServerApplication", LOG_LEVEL_INFO);

  // Create the nodes
  Ptr<Node> nT = CreateObject<Node> (); // node tramsittor
  Ptr<Node> nR1 = CreateObject<Node> (); // first node receiver
  Ptr<Node> nR2 = CreateObject<Node> (); // second node receiver
  Ptr<Node> nTd = CreateObject<Node> (); // "transmitting" diode
  Ptr<Node> nR1d = CreateObject<Node> (); // first "receiving" diode
  Ptr<Node> nR2d = CreateObject<Node> (); // second "receiving" diode

  Names::Add("nT", nT);
  Names::Add("nR1", nR1);
  Names::Add("nR2", nR2);
  Names::Add("nTd", nTd);
  Names::Add("nR1d", nR1d);
  Names::Add("nR2d", nR2d);

  NodeContainer nodesT = NodeContainer (nTd, nT);
  NodeContainer nodesR1 = NodeContainer (nR1d, nR1);
  NodeContainer nodesR2 = NodeContainer (nR2d, nR2);
  NodeContainer nTdnR1d = NodeContainer (nTd, nR1d);
  NodeContainer nTdnR2d = NodeContainer (nTd, nR2d);

  // Create the network link
  CsmaHelper csma;

  // Create the "netwrok interfaces" and add them to the appropriate nodes
  NetDeviceContainer sendNet;
  sendNet = csma.Install(nodesT);
  NetDeviceContainer reciveNet1;
  reciveNet1 = csma.Install(nodesR1);
  NetDeviceContainer reciveNet2;
  reciveNet2 = csma.Install(nodesR2);
  NetDeviceContainer netnTdnR1d;
  netnTdnR1d = csma.Install(nTdnR1d);
  NetDeviceContainer netnTdnR2d;
  netnTdnR2d = csma.Install(nTdnR2d);
  
  InternetStackHelper stack;
  stack.Install(nodesT); 
  stack.Install(nodesR1); 
  stack.Install(nodesR2); 
 
  // Give the different networks IP-adresses
  Ipv4AddressHelper address;
  // Transmitt network
  address.SetBase ("10.0.1.0" , "255.255.255.0");
  Ipv4InterfaceContainer interfaceT;
  interfaceT = address.Assign (sendNet);

  // Receive network
  address.SetBase ("10.0.2.0" , "255.255.255.0");
  Ipv4InterfaceContainer interfaceR1;
  interfaceR1 = address.Assign (reciveNet1);
  
  // Second Receive network
  address.SetBase ("10.0.3.0" , "255.255.255.0");
  Ipv4InterfaceContainer interfaceR2;
  interfaceR2 = address.Assign (reciveNet2);

  // Network to transmit from nT to nR1
  address.SetBase ("192.168.0.0", "255.255.255.252");
  Ipv4InterfaceContainer intnTdnR1d;
  intnTdnR1d = address.Assign (netnTdnR1d);
  
  // Network to transmit from nT to nR2
  address.SetBase ("192.168.0.4", "255.255.255.252");
  Ipv4InterfaceContainer intnTdnR2d;
  intnTdnR2d = address.Assign (netnTdnR2d);

  // Autopopulate and generate routing tables
  Ipv4GlobalRoutingHelper::PopulateRoutingTables();
 
  // Set variables that is used by the "services"
  float cStart = 1.0;
  float cStop = nPackets * 0.1 + 10;    // Add ten seconds buffer to ensure everything is transmitted
  float sStop = cStop;
  float pInterval = 0.1;
  uint32_t pSize = 1024;

  // Add randomnes to the interval so not everything is sent at the same time
  RngSeedManager::SetSeed (3);  // Changes seed from default of 1 to 3
  RngSeedManager::SetRun (7);   // Changes run number from default of 1 to 7
  Ptr<UniformRandomVariable> randPtr = CreateObject<UniformRandomVariable> ();
  
  // Configuration of the "test" application
  UdpEchoServerHelper echoServer (9);

  ApplicationContainer serverApps1 = echoServer.Install (nR1);
  serverApps1.Start (Seconds (cStart));
  serverApps1.Stop (Seconds (cStop));
  
  ApplicationContainer serverApps2 = echoServer.Install (nR2);
  serverApps2.Start (Seconds (cStart));
  serverApps2.Stop (Seconds (cStop));

  // Packets from nT to nR1 and nR2
  UdpEchoClientHelper echoClient1 (interfaceR1.GetAddress(1), 9);
  echoClient1.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient1.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient1.SetAttribute ("PacketSize", UintegerValue (pSize));
  
  UdpEchoClientHelper echoClient2 (interfaceR2.GetAddress(1), 9);
  echoClient2.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient2.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient2.SetAttribute ("PacketSize", UintegerValue (pSize));

  ApplicationContainer clientApps1 = echoClient1.Install (nT);
  clientApps1.Start (Seconds (cStart));
  clientApps1.Stop (Seconds (cStop));
  
  ApplicationContainer clientApps2 = echoClient2.Install (nT);
  clientApps2.Start (Seconds (cStart));
  clientApps2.Stop (Seconds (cStop));

  // Packets from nR1 to nR2
  UdpEchoClientHelper echoClient3 (interfaceR2.GetAddress(1), 9);
  echoClient3.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient3.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient3.SetAttribute ("PacketSize", UintegerValue (pSize));
  
  ApplicationContainer clientApps3 = echoClient3.Install (nR1);
  clientApps3.Start (Seconds (cStart));
  clientApps3.Stop (Seconds (cStop));
 
  // Traffic form nR2 to nR1
  UdpEchoClientHelper echoClient4 (interfaceR1.GetAddress(1), 9);
  echoClient4.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient4.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient4.SetAttribute ("PacketSize", UintegerValue (pSize));
  
  ApplicationContainer clientApps4 = echoClient4.Install (nR2);
  clientApps4.Start (Seconds (cStart));
  clientApps4.Stop (Seconds (cStop));
  
  ApplicationContainer extraTrafic [nClients * 4]; // Adding an aditional client at all nodes

  // Additional Packets from nT to nR1 and nR2
  for(uint32_t round = 0; round < nClients; round += 1){
    UdpEchoClientHelper traficClient1 (interfaceR1.GetAddress(1), 9);
    traficClient1.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    //echoClient1.SetAttribute ("Interval", TimeValue (Seconds (randPtr->GetValue(0.0006, 0.001))));
    traficClient1.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient1.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round] = traficClient1.Install (nT);
    extraTrafic[round].Start (Seconds (cStart));
    extraTrafic[round].Stop (Seconds (cStop));
    
    UdpEchoClientHelper traficClient2 (interfaceR2.GetAddress(1), 9);
    traficClient2.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    //echoClient2.SetAttribute ("Interval", TimeValue (Seconds (randPtr->GetValue(0.0006, 0.001))));
    traficClient2.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient2.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round + 1] = traficClient2.Install (nT);
    extraTrafic[round + 1].Start (Seconds (cStart));
    extraTrafic[round + 1].Stop (Seconds (cStop));

    UdpEchoClientHelper traficClient3 (interfaceR1.GetAddress(1), 9);
    traficClient3.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    traficClient3.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient3.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round + 2] = traficClient3.Install (nR2);
    extraTrafic[round + 2].Start (Seconds (cStart));
    extraTrafic[round + 2].Stop (Seconds (cStop));
    
    UdpEchoClientHelper traficClient4 (interfaceR2.GetAddress(1), 9);
    traficClient4.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    traficClient4.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient4.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round + 3] = traficClient2.Install (nR1);
    extraTrafic[round + 3].Start (Seconds (cStart));
    extraTrafic[round + 3].Stop (Seconds (cStop));
  }
  
  // Print all the routing tables for debugging
  Ipv4GlobalRoutingHelper printRouting;
  Ptr<OutputStreamWrapper> routingStream = Create<OutputStreamWrapper> ("test-network-no-diodes.routes", std::ios::out);
  printRouting.PrintRoutingTableAllAt (Seconds (2), routingStream);
	
  // Packet dump of send network 
  csma.EnablePcap ("test-network-no-diodes", sendNet.Get (1), true);
  csma.EnablePcap ("test-network-no-diodes", sendNet.Get (0), true);
  // Packet dump of receive network
  csma.EnablePcap ("test-network-no-diodes", reciveNet1.Get (0), true);
  csma.EnablePcap ("test-network-no-diodes", reciveNet1.Get (1), true);
  csma.EnablePcap ("test-network-no-diodes", reciveNet2.Get (0), true);
  csma.EnablePcap ("test-network-no-diodes", reciveNet2.Get (1), true);
  csma.EnablePcap ("test-network-no-diodes", netnTdnR1d.Get (0), true);
  csma.EnablePcap ("test-network-no-diodes", netnTdnR1d.Get (1), true);
  csma.EnablePcap ("test-network-no-diodes", netnTdnR2d.Get (0), true);
  csma.EnablePcap ("test-network-no-diodes", netnTdnR2d.Get (1), true);
  
  // Flow monitor
  Ptr<FlowMonitor> flowMonitor;
  FlowMonitorHelper flowHelper;
  flowMonitor = flowHelper.InstallAll();
  
  Simulator::Stop (Seconds (sStop));
  Simulator::Run ();
  Simulator::Destroy ();
  
	flowMonitor->SerializeToXmlFile("test-network-no-diodes.xml", true, true);
  return 0;
}
