/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* Overview of the network 
 *
 *             nT
 *         -<-  |  -<-
 *        /   \ | /   \
 * nR1-nR1d    nTd     nR2d-nR2
 *        \   /   \   /
 *         ->-     ->-
 */

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/internet-module.h"
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/csma-module.h"
#include "ns3/arp-cache.h"
#include "ns3/ipv4-static-routing-helper.h"
#include "ns3/netanim-module.h"
#include "ns3/flow-monitor-helper.h"


// Normal C++ stuff
#include <iostream>
#include <fstream>
#include <string>

using namespace ns3;
using std::cout;

NS_LOG_COMPONENT_DEFINE ("NetworkDiodeTest");


// A function to create a data diode connection
void
CreateDiode (Ptr<Node> sender, 
    Ptr<Node> receiver,
    char const* adress,
    char const* subnetMask,
    char const* baseAdr,
    Ipv4StaticRoutingHelper* ipv4RoutingHelper,
    Ipv4Address destAdr,
    Ipv4Mask destMask,
    bool pcap=false
    )
{
  // Create the network link
  CsmaHelper csma;

  NodeContainer nodes = NodeContainer (sender, receiver);

  // Create the "network interfaces" and add them to the appropriate nodes
  NetDeviceContainer diodes;
  diodes = csma.Install(nodes);
  
  // Configure the interfaces as data diodes
  Ptr<CsmaNetDevice> diodeS = DynamicCast<CsmaNetDevice> (diodes.Get(0));
  diodeS->SetReceiveEnable (false);
  diodeS->SetSendEnable (true);
  
  Ptr<CsmaNetDevice> diodeR = DynamicCast<CsmaNetDevice> (diodes.Get(1));
  diodeR->SetReceiveEnable (true);
  diodeR->SetSendEnable (false);
 
  // Initalize Internet if needed
  InternetStackHelper stack;
  // Ensure that the node does not already have an initialized IP stack
  if (sender->GetObject<Ipv4> () != 0)
  {
    NS_LOG_INFO ("Sender already has an initialized IP stack");
  }
  else 
  {
    stack.Install(sender); 
  }
  if (receiver->GetObject<Ipv4> () != 0)
  {
    NS_LOG_INFO ("Receiver already has an initialized IP stack");
  }
  else 
  {
    stack.Install(receiver); 
  }
  
  // Set IP addresses
  Ipv4AddressHelper address;
  address.SetBase (adress, subnetMask, baseAdr); 
  Ipv4InterfaceContainer interfacesDiodes;
  interfacesDiodes = address.Assign (diodes);
  
  // Variables used for static routing  
  Ipv4Address destIntAdress = interfacesDiodes.GetAddress(1);
  uint32_t numInterface = diodeS->GetIfIndex();

  Ptr<Ipv4> ipv4S = sender->GetObject<Ipv4> ();

  // Use static routing for the diodes, hopefully allowing for "loops"
  Ptr<Ipv4StaticRouting> staticRouteT = ipv4RoutingHelper->GetStaticRouting (ipv4S);
  staticRouteT->AddNetworkRouteTo (destAdr, destMask, destIntAdress, numInterface);
  // Comment out to disable default routes
  staticRouteT->SetDefaultRoute (destIntAdress, numInterface);
  
  // Manually fill the ARP cache of the transmit node
  Ptr<ArpCache> arpT = CreateObject<ArpCache> ();
  arpT->SetAliveTimeout (Seconds (3600 * 24)); // Keep the ARP table entry for one day...
   
  ArpCache::Entry * entry = arpT->Add (interfacesDiodes.GetAddress(1));
  entry->SetMacAddress(Mac48Address::ConvertFrom(diodeR->GetAddress()));
  entry->MarkPermanent ();

  // Add the cache to the transmit node
  std::pair<Ptr<Ipv4>, uint32_t> returnValue = interfacesDiodes.Get(0);
  Ptr<Ipv4> ipv4 = returnValue.first;
  uint32_t index = returnValue.second;
  Ptr<Ipv4Interface> diodeT = ipv4->GetObject<Ipv4L3Protocol> ()->GetInterface (index);
  arpT->SetDevice(diodeS, diodeT);
  diodeT->SetAttribute("ArpCache", PointerValue(arpT));
  
  // Enable packet trace on diode interfaces
  if (pcap)
  {
    csma.EnablePcap ("diode-send", diodeS, true);
    csma.EnablePcap ("diode-receive", diodeR, true);
  }
}

int
main (int argc, char *argv[])
{
  uint32_t nPackets = 100;
  uint32_t nClients = 10;

  CommandLine cmd;
  cmd.AddValue("nPackets", "Number of packets to echo", nPackets);
  cmd.AddValue("nClients", "Number of extra clients to run", nClients);
  cmd.Parse (argc, argv);
  
  Time::SetResolution (Time::NS);
  LogComponentEnable ("UdpEchoClientApplication", LOG_LEVEL_INFO);
  LogComponentEnable ("UdpEchoServerApplication", LOG_LEVEL_INFO);

  // Create the nodes
  Ptr<Node> nT = CreateObject<Node> (); // node tramsittor
  Ptr<Node> nT2 = CreateObject<Node> (); // node tramsittor
  Ptr<Node> nR1 = CreateObject<Node> (); // first node receiver
  Ptr<Node> nR2 = CreateObject<Node> (); // second node receiver
  Ptr<Node> nTd = CreateObject<Node> (); // "transmitting" diode
  Ptr<Node> nR1d = CreateObject<Node> (); // first "receiving" diode
  Ptr<Node> nR2d = CreateObject<Node> (); // second "receiving" diode

  Names::Add("nT", nT);
  Names::Add("nT2", nT2);
  Names::Add("nR1", nR1);
  Names::Add("nR2", nR2);
  Names::Add("nTd", nTd);
  Names::Add("nR1d", nR1d);
  Names::Add("nR2d", nR2d);

  NodeContainer nodesT = NodeContainer (nTd, nT);
  NodeContainer nodesR1 = NodeContainer (nR1d, nR1);
  NodeContainer nodesR2 = NodeContainer (nR2d, nR2);

  // Create the network link
  CsmaHelper csma;

  // Create the "network interfaces" and add them to the appropriate nodes
  NetDeviceContainer sendNet;
  sendNet = csma.Install(nodesT);
  NetDeviceContainer reciveNet1;
  reciveNet1 = csma.Install(nodesR1);
  NetDeviceContainer reciveNet2;
  reciveNet2 = csma.Install(nodesR2);
  
  InternetStackHelper stack;
  stack.Install(nodesT); 
  stack.Install(nodesR1); 
  stack.Install(nodesR2); 
 
  // Assign fixed IP-adresses to the networks
  Ipv4AddressHelper address;
  // Transmitt network
  address.SetBase ("10.0.1.0" , "255.255.255.0");
  Ipv4InterfaceContainer interfaceT;
  interfaceT = address.Assign (sendNet);

  // Receive network
  address.SetBase ("10.0.2.0" , "255.255.255.0");
  Ipv4InterfaceContainer interfaceR1;
  interfaceR1 = address.Assign (reciveNet1);
  
  // Second Receive network
  address.SetBase ("10.0.3.0" , "255.255.255.0");
  Ipv4InterfaceContainer interfaceR2;
  interfaceR2 = address.Assign (reciveNet2);
 
  // Enable static routing
  Ipv4StaticRoutingHelper ipv4RoutingHelper; 
  
  // Configure nTd, nR1d, and nR2d to function as diodes
  CreateDiode(nTd, nR1d, "192.168.0.0", "255.255.255.0", "0.0.0.1", &ipv4RoutingHelper, "10.0.2.0", "/24", true);
  CreateDiode(nR1d, nTd, "192.168.0.0", "255.255.255.0", "0.0.0.3", &ipv4RoutingHelper, "10.0.1.0", "/24", true);
  CreateDiode(nTd, nR2d, "192.168.0.0", "255.255.255.0", "0.0.0.5", &ipv4RoutingHelper, "10.0.3.0", "/24", true);
  CreateDiode(nR2d, nTd, "192.168.0.0", "255.255.255.0", "0.0.0.7", &ipv4RoutingHelper, "10.0.1.0", "/24", true);
  
  // Static routing
  Ptr<Ipv4> ipv4nT = nT->GetObject<Ipv4> ();
  Ptr<Ipv4> ipv4nR1 = nR1->GetObject<Ipv4> ();
  Ptr<Ipv4> ipv4nR2 = nR2->GetObject<Ipv4> ();
  Ptr<Ipv4> ipv4nTd = nTd->GetObject<Ipv4> ();
  Ptr<Ipv4> ipv4nR1d = nR1d->GetObject<Ipv4> ();
  Ptr<Ipv4> ipv4nR2d = nR2d->GetObject<Ipv4> ();

  // Use static routing for the diodes hopefully allowing for "loops"
  // The primary goal is to populate the adress of the other side of the diode
  Ptr<Ipv4StaticRouting> staticRoutenT = ipv4RoutingHelper.GetStaticRouting (ipv4nT);
  staticRoutenT->AddNetworkRouteTo (Ipv4Address ("10.0.2.0"), Ipv4Mask ("/24"), Ipv4Address ("10.0.1.1"), 1);
  staticRoutenT->AddNetworkRouteTo (Ipv4Address ("10.0.3.0"), Ipv4Mask ("/24"), Ipv4Address ("10.0.1.1"), 1);

  Ptr<Ipv4StaticRouting> staticRoutenR1 = ipv4RoutingHelper.GetStaticRouting (ipv4nR1);
  staticRoutenR1->AddNetworkRouteTo (Ipv4Address ("10.0.1.0"), Ipv4Mask ("/24"), Ipv4Address("10.0.2.1"), 1);
  staticRoutenR1->SetDefaultRoute (Ipv4Address("10.0.2.1"), 1);
  
  Ptr<Ipv4StaticRouting> staticRoutenR2 = ipv4RoutingHelper.GetStaticRouting (ipv4nR2);
  staticRoutenR2->AddNetworkRouteTo (Ipv4Address ("10.0.1.0"), Ipv4Mask ("/24"), Ipv4Address("10.0.3.1"), 1);
  staticRoutenR2->SetDefaultRoute (Ipv4Address("10.0.3.1"), 1);
  
  Ptr<Ipv4StaticRouting> staticRoutenTd = ipv4RoutingHelper.GetStaticRouting (ipv4nTd);

  Ptr<Ipv4StaticRouting> staticRoutenR1d = ipv4RoutingHelper.GetStaticRouting (ipv4nR1d);
  // Comment in to enable blackholing of nR2 on nR1d
  // staticRoutenR1d->AddNetworkRouteTo (Ipv4Address ("10.0.3.0"), Ipv4Mask ("/24"), 2);
  
  Ptr<Ipv4StaticRouting> staticRoutenR2d = ipv4RoutingHelper.GetStaticRouting (ipv4nR2d);
  // Comment in to enable blackholing of nR1 on nR2d
  // staticRoutenR2d->AddNetworkRouteTo (Ipv4Address ("10.0.2.0"), Ipv4Mask ("/24"), 2);
  
  // Set variables that is used by the "services"
  float cStart = 1.0;
  float cStop = nPackets * 0.1 + 10;    // Add ten seconds buffer to ensure everything is transmitted
  float sStop = cStop;
  float pInterval = 0.01;
  uint32_t pSize = 1024;

  // Add randomnes to the inter-packet interval to prevent each node
  // from transmitting at the same time
  RngSeedManager::SetSeed (3);  // Changes seed from default of 1 to 3
  RngSeedManager::SetRun (7);   // Changes run number from default of 1 to 7
  Ptr<UniformRandomVariable> randPtr = CreateObject<UniformRandomVariable> ();
  
  // Configuration of the "test" application
  UdpEchoServerHelper echoServer (9);

  ApplicationContainer serverApps1 = echoServer.Install (nR1);
  serverApps1.Start (Seconds (cStart));
  serverApps1.Stop (Seconds (cStop));
  
  ApplicationContainer serverApps2 = echoServer.Install (nR2);
  serverApps2.Start (Seconds (cStart));
  serverApps2.Stop (Seconds (cStop));

  // Packets from nT to nR1 and nR2
  UdpEchoClientHelper echoClient1 (interfaceR1.GetAddress(1), 9);
  echoClient1.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient1.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient1.SetAttribute ("PacketSize", UintegerValue (pSize));
  
  UdpEchoClientHelper echoClient2 (interfaceR2.GetAddress(1), 9);
  echoClient2.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient2.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient2.SetAttribute ("PacketSize", UintegerValue (pSize));

  ApplicationContainer clientApps1 = echoClient1.Install (nT);
  clientApps1.Start (Seconds (cStart));
  clientApps1.Stop (Seconds (cStop));
  
  ApplicationContainer clientApps2 = echoClient2.Install (nT);
  clientApps2.Start (Seconds (cStart));
  clientApps2.Stop (Seconds (cStop));

  // Packets from nR1 to nR2
  UdpEchoClientHelper echoClient3 (interfaceR2.GetAddress(1), 9);
  echoClient3.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient3.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient3.SetAttribute ("PacketSize", UintegerValue (pSize));
  
  ApplicationContainer clientApps3 = echoClient3.Install (nR1);
  clientApps3.Start (Seconds (cStart));
  clientApps3.Stop (Seconds (cStop));
 
  // Traffic form nR2 to nR1
  UdpEchoClientHelper echoClient4 (interfaceR1.GetAddress(1), 9);
  echoClient4.SetAttribute ("MaxPackets", UintegerValue (nPackets));
  echoClient4.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
  echoClient4.SetAttribute ("PacketSize", UintegerValue (pSize));
  
  ApplicationContainer clientApps4 = echoClient4.Install (nR2);
  clientApps4.Start (Seconds (cStart));
  clientApps4.Stop (Seconds (cStop));

  ApplicationContainer extraTrafic [nClients * 4]; // Adding an aditional client at all nodes

  // Additional traffic between the nodes
  for(uint32_t round = 0; round < nClients; round += 1){
    UdpEchoClientHelper traficClient1 (interfaceR1.GetAddress(1), 9);
    traficClient1.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    traficClient1.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient1.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round] = traficClient1.Install (nT);
    extraTrafic[round].Start (Seconds (cStart));
    extraTrafic[round].Stop (Seconds (cStop));
    
    UdpEchoClientHelper traficClient2 (interfaceR2.GetAddress(1), 9);
    traficClient2.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    traficClient2.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient2.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round + 1] = traficClient2.Install (nT);
    extraTrafic[round + 1].Start (Seconds (cStart));
    extraTrafic[round + 1].Stop (Seconds (cStop));

    UdpEchoClientHelper traficClient3 (interfaceR1.GetAddress(1), 9);
    traficClient3.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    traficClient3.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient3.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round + 2] = traficClient3.Install (nR2);
    extraTrafic[round + 2].Start (Seconds (cStart));
    extraTrafic[round + 2].Stop (Seconds (cStop));
    
    UdpEchoClientHelper traficClient4 (interfaceR2.GetAddress(1), 9);
    traficClient4.SetAttribute ("MaxPackets", UintegerValue (nPackets));
    traficClient4.SetAttribute ("Interval", TimeValue (Seconds (pInterval)));
    traficClient4.SetAttribute ("PacketSize", UintegerValue (pSize));
    
    extraTrafic[round + 3] = traficClient2.Install (nR1);
    extraTrafic[round + 3].Start (Seconds (cStart));
    extraTrafic[round + 3].Stop (Seconds (cStop));
  }



  // Print all the routing tables for debugging
  Ipv4GlobalRoutingHelper printRouting;
  Ptr<OutputStreamWrapper> routingStream = Create<OutputStreamWrapper> ("test-network.routes", std::ios::out);
  printRouting.PrintRoutingTableAllAt (Seconds (2), routingStream);
	
  // Packet dump of receive network
  csma.EnablePcap ("test-network", reciveNet1.Get (1), true);
  csma.EnablePcap ("test-network", reciveNet2.Get (1), true);
  // Flow monitor
  Ptr<FlowMonitor> flowMonitor;
  FlowMonitorHelper flowHelper;
  flowMonitor = flowHelper.InstallAll();

  Simulator::Stop (Seconds (sStop));
  Simulator::Run ();
  Simulator::Destroy ();

  flowMonitor->SerializeToXmlFile("test-network.xml", true, true);
  return 0;
}
